"""
Provides a framework for creating unit tests for projects based on 
py-gtfs-loader. All test suites must follow a common structure:

test_something/ - A test case must begin with test_.
    input/ - An input GTFS feed to be loaded into the tool.
    expected_default/ - The output the tool is expected to produce.
    expected_abc/ - The same input data can be used for several related tests with different expected outputs.
        A test runner can use `tag='abc'` to compare against the `abc` data for example.

See tests/ in this repo for a simple example of the framework in use.
"""


from pathlib import Path
import tempfile
import shutil
import os


def init(caller_location):
    """
    Prepare to run tests by setting up paths and creating temporary directories.
    """

    global TEST_DIR
    global WORK_DIR

    TEST_DIR = Path(caller_location).parent.resolve()
    WORK_DIR = TEST_DIR / '.work'
    WORK_DIR.mkdir(exist_ok=True)


def find_tests(tag='default'):
    """
    Discover directories containing test feeds with a certain tag.
    """

    test_dirs = []
    for dirent in TEST_DIR.iterdir():
        if dirent.is_dir() and dirent.name.startswith('test_'):
            if not any(dirent.glob(f'expected_{tag}')):
                continue

            test_dirs.append(dirent)

    return test_dirs


def create_test_data(feed_dir):
    """
    Combine the base files from the feed (common to all tests) with the input
    files for a particular test case, in a temporary directory. The name of 
    this directory is returned.
    """

    work_dir = Path(tempfile.mkdtemp(prefix='', dir=WORK_DIR))

    for filename in (TEST_DIR / 'base').iterdir():
        shutil.copy2(filename, work_dir / filename.name)

    for filename in (feed_dir / 'input').iterdir():
        shutil.copy2(filename, work_dir / filename.name)

    print(f'Testing feed in {work_dir}')
    return work_dir


def check_expected_output(feed_dir, work_dir, tag='default'):
    """
    Check whether work_dir contains the expected output. If successful, work_dir
    will be removed at the end of the test.
    """

    expected_dir = feed_dir / f'expected_{tag}'
    
    for expected_filename in (feed_dir / expected_dir).iterdir():
        actual_filename = work_dir / expected_filename.name
        check_file(expected_filename, actual_filename)

    shutil.rmtree(work_dir)


def check_file(expected_filename, actual_filename):
    """
    Check that two files are identical and print a readable diff if not.
    Works best on CSV files
    """

    with open(actual_filename) as actual_fp:
        actual_text = [line.strip() for line in actual_fp.readlines()]

    with open(expected_filename) as expected_fp:
        expected_text = [line.strip() for line in expected_fp.readlines()] 

    check_snapshot_update(expected_filename, expected_text, actual_text)
    assert actual_text == expected_text


def check_snapshot_update(expected_filename, expected_text, actual_text):
    """
    If UPDATE_SNAPSHOTS=1 and the test fails because the expected output files
    are obsolete, update the expected files to match the test.
    """

    if not int(os.environ.get('UPDATE_SNAPSHOTS', 0)):
        return

    if actual_text == expected_text:
        return

    with open(expected_filename, 'w', encoding='utf-8') as f:
        f.write('\n'.join(actual_text))
        print(f'Updated {expected_filename}')



import functools
import enum
import json
from datetime import datetime
from typing import Any

from .schema_classes import Schema, SchemaCollection


class GTFSTime(int):
    # GTFS allows times exceeding 23:59:59 as it is simpler to describe night
    # services that way in many cases. A trip could theoretically be shifted
    # forward an arbitrary number of days using this notation, but we block it
    # as it just creates confusion.
    MAX_HOUR_REPRESENTATION = 36
    def __new__(cls, time_str):
        if isinstance(time_str, int):
            return super().__new__(cls, time_str)

        if time_str == '':
            return super().__new__(cls, -1)

        h, m, s = time_str.split(':')

        if int(h) > GTFSTime.MAX_HOUR_REPRESENTATION:
            raise ValueError(
                f'Refusing to consider a service day longer than {GTFSTime.MAX_HOUR_REPRESENTATION} hours'
            )

        return super().__new__(cls, 3600 * int(h) + 60 * int(m) + int(s))

    def __str__(self):
        if self == -1:
            return ''

        hours, rem = divmod(self, 3600)
        mins, secs = divmod(rem, 60)
        return '%02d:%02d:%02d' % (hours, mins, secs)

    def __add__(self, other):
        return GTFSTime(super().__add__(other))

    def __sub__(self, other):
        return GTFSTime(super().__sub__(other))

class GTFSDate(datetime):

    def __new__(cls, *args, **kwargs):
        if len(args) != 1 or kwargs:
            return super().__new__(cls, *args, **kwargs)

        iso_str = args[0]
        if isinstance(iso_str, datetime):
            return super().__new__(cls,
                                   year=iso_str.year,
                                   month=iso_str.month,
                                   day=iso_str.day)

        if not iso_str:
            raise ValueError('Invalid date: empty')
        try:
            return cls.strptime(iso_str, '%Y-%m-%d')
        except ValueError:
            return cls.strptime(iso_str, '%Y%m%d')

    def __repr__(self):
        return self.strftime('%Y%m%d')

    def __str__(self):
        return repr(self)


class EntityDict(dict):

    def __init__(self, fields, values=None):
        super().__init__(values if values else [])
        self._resolved_fields = fields


class Entity:

    def __init__(self, **kwargs):
        self._gtfs = None
        default_init = {
            k: v
            for k, v in self.__class__.__dict__.items()
            if Entity._is_field(k, v)
        }
        self.__dict__.update(default_init)
        self.__dict__.update(kwargs)

    @staticmethod
    def _is_field(k, v):
        if callable(v) or isinstance(v, functools.cached_property):
            return False

        return not k.startswith('_')

    def __getitem__(self, item):
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __delitem__(self, key):
        del self.__dict__[key]

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def __repr__(self):
        filtered_dict = {
            k: v for k, v in self.__dict__.items() if k not in {'_gtfs'}
        }
        return f'{self.__class__.__name__} {repr(filtered_dict)}'

    def clone(self, **overrides):
        merged = {
            k: v for k, v in self.__dict__.items() if Entity._is_field(k, v)
        }
        merged.update(overrides)
        return self.__class__(**merged)


@functools.singledispatch
def serialize(value: Any):
    if isinstance(value, list):
        # Remove spaces after commas for compactness
        return json.dumps(value,separators=(',', ':')) 
    return str(value)


@serialize.register
def _(value: enum.IntEnum):
    return str(int(value))


@serialize.register
def _(value: bool):
    return str(int(value))


@serialize.register
def _(value: None):
    return ''


class Properties(Entity):
    _schema = Schema()

    stop_name: str = ''
    stop_desc: str = ''
    zone_id: str = ''
    stop_url: str = ''


class Geometry(Entity):
    _schema = Schema()

    type: str
    coordinates: Any


class Feature(Entity):
    _schema = Schema()
    id: str
    type: str = "Feature"  # Need to be asserted
    properties: Properties
    geometry: Geometry


SCHEMA_COLLECTION = SchemaCollection(Properties, Geometry, Feature)



import csv
import enum
import json
import os
import shutil
import typing
from pathlib import Path
from . import schema_classes, types, schema


class ParseError(ValueError):
    pass

def get_files(files):
    return schema.FileCollection(*(schema.GTFS_FILENAMES[f] for f in files)).values()


def load(gtfs_dir, sorted_read=False, files=None, verbose=True, itineraries=False):
    gtfs_dir = Path(gtfs_dir)
    gtfs = types.Entity()

    files_to_load = get_files(files) if files else schema.GTFS_SUBSET_SCHEMA_ITINERARIES.values() if itineraries else schema.GTFS_SUBSET_SCHEMA.values()

    for file_schema in files_to_load:
        if verbose:
            print(f'Loading {file_schema.name}')
        filepath = gtfs_dir / file_schema.filename
        gtfs[file_schema.name] = types.EntityDict(
            file_schema.get_declared_fields())

        if not filepath.exists():
            if file_schema.required:
                raise ParseError(
                    f'{file_schema.filename}: required file is missing')
            else:
                continue

        if file_schema.fileType is schema_classes.FileType.CSV:
            load_csv(gtfs, filepath, file_schema, sorted_read=(True if file_schema.name == 'stop_times' or file_schema.name == 'shapes' else sorted_read))
        elif file_schema.fileType is schema_classes.FileType.GEOJSON:
            load_json(gtfs, filepath, file_schema)

    return gtfs


def load_csv(gtfs, filepath, file_schema, sorted_read=False):
    with open(filepath, 'r', encoding='utf-8-sig') as f:
        reader = csv.reader(f, skipinitialspace=True)
        header_row = next(reader, None)
        if not header_row:
            if file_schema.required:
                raise ParseError(
                    f'{file_schema.filename}: required file is empty')
            else:
                return

        resolved_fields = merge_header_and_declared_fields(
            file_schema, header_row)
        entities = {}
        for entity in parse_rows(gtfs, file_schema, resolved_fields,
                                 header_row, reader):
            index_entity(file_schema, entities, entity)

        if sorted_read:
            processed_entities = sorted_entities(file_schema, entities)
        else:
            processed_entities = entities.items()

        gtfs[file_schema.name] = types.EntityDict(fields=resolved_fields,
                                                  values=processed_entities)


def load_json(gtfs, filepath, file_schema):
    with open(filepath, 'r', encoding='utf-8-sig') as f:
        json_data = json.load(f)

        gtfs[file_schema.name] = visit_json(json_data, file_schema.class_def())


def visit_json(json_data, expected_type, type_config=None):
    if isinstance(json_data, dict):
        file_schema = expected_type._schema
        declared_fields = file_schema.get_declared_fields()
        output = file_schema.class_def()

        for name, config in declared_fields.items():
            value = json_data.get(name, None)

            if config.required and value is None:
                raise ParseError(
                    f'{expected_type.__name__} missing required field {name}')

            output[name] = visit_json(value, config.type, config)

        return output
    elif isinstance(json_data, list):
        output = []
        for value in json_data:
            inner_type = get_inner_type(expected_type)
            output.append(visit_json(value, inner_type))

        return output
    else:
        if type_config is not None and not type_config.required and json_data == None:
            return type_config.default

        if expected_type is typing.Any:
            return json_data

        return expected_type(json_data)


def merge_header_and_declared_fields(file_schema, header_row):
    declared_fields = file_schema.get_declared_fields()
    fields = {}

    for name in header_row:
        fields[name] = declared_fields.get(name) or schema_classes.Field(
            str, False, '')

    for name, config in declared_fields.items():
        if config.required and name not in fields:
            raise ParseError(
                f'{file_schema.filename}:1: missing required field {name}')

        fields.setdefault(name, config)

    return fields


def parse_rows(gtfs, file_schema, fields, header_row, reader):
    for lineno, row in enumerate(reader, 2):
        if len(row) == 0:
            continue  # empty row, just skip it

        entity = file_schema.class_def()
        entity._gtfs = gtfs

        for name, value in zip(header_row, row):
            config = fields[name]
            if config.required and not value:
                raise ParseError(
                    f'{file_schema.filename}:{lineno}: required field {name} is empty'
                )

            entity[name] = validate(
                config,
                value,
                context_fn=lambda:
                f'{file_schema.filename}:{lineno} field {name} = {repr(value)}')

        yield entity


def validate(config, value, context_fn):
    try:
        return convert(config, value)
    except Exception as exc:
        raise ParseError(f'{context_fn()}: {exc.args[0]}') from None


def convert(config, value):
    if not config.required and value == '':
        return config.default

    # Lists are stringified as JSON in csv.
    if typing.get_origin(config.type) is list:
        return list(json.loads(value))
    
    config_type = get_inner_type(config.type)
    if issubclass(config_type, enum.IntEnum):
        return config_type(int(value))

    if config_type is bool:
        return bool(int(value))

    return config_type(value)


def get_inner_type(config_type):
    if typing.get_origin(config_type) is list:
        return typing.get_args(config_type)[0]

    if typing.get_origin(config_type) is not typing.Union:
        return config_type

    variants = typing.get_args(config_type)
    if len(variants) != 2:
        raise ValueError("Misconfigured type definition")

    for variant in variants:
        if not isinstance(None, variant):
            return variant

    raise ValueError("Misconfigured type definition")


def index_entity(file_schema, entities, entity):
    key = entity[file_schema.id]
    if not file_schema.group_id:
        entities[key] = entity
        return

    group_key = entity[file_schema.group_id]
    if file_schema.inner_dict:
        entities.setdefault(key, {})[group_key] = entity
    else:
        entities.setdefault(key, []).append(entity)


def sorted_entities(file_schema, entities):
    if file_schema.group_id:
        if file_schema.inner_dict:
            for group_key, group in entities.items():
                entities[group_key] = dict(
                    sorted(group.items(), key=lambda kv: kv[0]))
        else:
            for group in entities.values():
                group.sort(key=lambda entity: entity[file_schema.group_id])

    return sorted(entities.items(), key=lambda kv: kv[0])


def patch(gtfs, gtfs_in_dir, gtfs_out_dir, files=None, sorted_output=False, verbose=True, itineraries=False):
    gtfs_in_dir = Path(gtfs_in_dir)
    gtfs_out_dir = Path(gtfs_out_dir)
    gtfs_out_dir.mkdir(parents=True, exist_ok=True)

    for original_filename in gtfs_in_dir.iterdir():
        try:
            shutil.copy2(original_filename,
                         gtfs_out_dir / original_filename.name)
        except shutil.SameFileError:
            pass  # No need to copy if we're working in-place

    files_to_patch = get_files(files) if files else schema.GTFS_SUBSET_SCHEMA_ITINERARIES.values() if itineraries else schema.GTFS_SUBSET_SCHEMA.values()

    for file_schema in files_to_patch:
        if verbose:
            print(f'Writing {file_schema.name}')
        entities = gtfs.get(file_schema.name)
        if not entities:
            (gtfs_out_dir / file_schema.filename).unlink(missing_ok=True)
            continue

        if file_schema.fileType is schema_classes.FileType.CSV:
            save_csv(file_schema, entities, gtfs_out_dir, sorted_output)
        elif file_schema.fileType is schema_classes.FileType.GEOJSON:
            save_json(file_schema, entities, gtfs_out_dir)


def save_csv(file_schema, entities, gtfs_out_dir, sorted_output=False):
    if sorted_output:
        processed_entities = dict(sorted_entities(file_schema, entities))
    else:
        processed_entities = entities.copy()

    flat_entities = flatten_entities(file_schema, processed_entities)
    fields = entities._resolved_fields

    with open(gtfs_out_dir / (file_schema.filename), 'w', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(fields.keys())
        for entity in flat_entities:
            writer.writerow(
                types.serialize(entity.get(name, '')) for name in fields)


def save_json(file_schema, entities, gtfs_out_dir):
    with open(gtfs_out_dir / (file_schema.filename), 'w', encoding='utf-8') as f:
        f.write(json.dumps(entities, indent=4, default=vars))


def flatten_entities(file_schema, entities):
    if file_schema.group_id:
        flat_entities = []
        for entity_list in entities.values():
            flat_entities.extend(
                entity_list.values() if file_schema.inner_dict else entity_list)
        return flat_entities
    else:
        return entities.values()


def clone(entities, key, new_key):
    if key not in entities:
        return

    entries = entities[key]
    if isinstance(entries, list):
        entities[new_key] = [
            clone_and_index(entity, new_key) for entity in entries
        ]
    else:
        entities[new_key] = clone_and_index(entries, new_key)

    return entities[new_key]


def clone_and_index(entity, new_key):
    field_schema = entity.__class__._schema
    new_entity = entity.clone()
    new_entity[field_schema.id] = new_key
    return new_entity



from collections import namedtuple
from enum import IntEnum

Field = namedtuple('Field', ('type', 'required', 'default'))


class FileType(IntEnum):
    CSV = 0
    GEOJSON = 1


class Schema:
    def __init__(self, required=True, group_id=None, inner_dict=False):
        # If this file is absent from the GTFS directory, should process abort?
        self.required = required

        # If set, multiple entities may share the same ID. The group key is used to provide a deterministic order among
        # such entities
        self.group_id = group_id

        # If true, then the group key should be a dictionary key at the second level
        self.inner_dict = inner_dict

        # Will be set to point to the class defining this file
        self.class_def = None

    def get_declared_fields(self):
        return {
            k: Field(type=v,
                     required=k not in self.class_def.__dict__,
                     default=self.class_def.__dict__.get(k))
            for k, v in self.class_def.__annotations__.items()
        }


class File(Schema):

    def __init__(self,
                 id,
                 name,
                 fileType,
                 filename=None,
                 required=True,
                 group_id=None,
                 inner_dict=False):

        super().__init__(required, group_id, inner_dict)
        # Primary key for this file, serves as the key of the generated dict
        self.id = id

        # Objects will appear under gtfs."name"
        self.name = name

        # Specify if the file is a csv or a json
        self.fileType = fileType

        # Objects will be read from "filename".txt
        # If unset, the entity name shall be used instead
        self.filename = filename if filename else name + \
            File._get_file_ext(fileType)

    @staticmethod
    def _get_file_ext(fileType):
        if fileType is FileType.CSV:
            return ".txt"

        if fileType is FileType.GEOJSON:
            return ".geojson"


class FileCollection:

    def __init__(self, *args):
        self.entities = {}
        for file in args:
            file._schema.class_def = file
            self.entities[file._schema.filename] = file._schema

    def keys(self):
        return self.entities.keys()

    def values(self):
        return self.entities.values()

    def __getitem__(self, item):
        return self.entities[item]


class SchemaCollection:

    def __init__(self, *args):
        for file in args:
            file._schema.class_def = file




from math import *


class LatLon:
    # Sources:
    # http://www.movable-type.co.uk/scripts/gis-faq-5.1.html
    # http://www.movable-type.co.uk/scripts/latlong.html
    EARTH_RADIUS_M = 6_371_009.0
    __slots__ = ('lat', 'lon')

    def __init__(self, lat, lon, unit=degrees):
        self.lat = radians(lat) if unit == degrees else lat
        self.lon = radians(lon) if unit == degrees else lon

    def geojson(self):
        return f'[{degrees(self.lon)}, {degrees(self.lat)}]'

    def __repr__(self):
        return repr((degrees(self.lat), degrees(self.lon)))

    def angular_distance_to(self, other):
        d_lat = other.lat - self.lat
        d_lon = other.lon - self.lon
        a = sin(
            d_lat / 2)**2 + cos(self.lat) * cos(other.lat) * sin(d_lon / 2)**2
        return 2 * asin(sqrt(a))

    def distance_to(self, other):
        return LatLon.EARTH_RADIUS_M * self.angular_distance_to(other)

    def bearing_to(self, other):
        y = sin(other.lon - self.lon) * cos(other.lat)
        x = cos(self.lat) * sin(other.lat) - sin(self.lat) * cos(
            other.lat) * cos(other.lon - self.lon)
        return atan2(y, x)

    def add_bearing_and_angular_distance(self, bearing, dist):
        lat = asin(
            sin(self.lat) * cos(dist) +
            cos(self.lat) * sin(dist) * cos(bearing))
        lon = self.lon \
            + atan2(sin(bearing)*sin(dist)*cos(self.lat),
                    cos(dist) - sin(self.lat)*sin(lat))

        return LatLon(lat, lon, unit=radians)

    def distance_to_segment(x, l1, l2):
        """
                     x
                     |
                     |
        l1 ---------lx------------ l2
        """

        d_l1_x, t_l1_x = l1.angular_distance_to(x), l1.bearing_to(x)
        d_l1_l2, t_l1_l2 = l1.angular_distance_to(l2), l1.bearing_to(l2)
        d_l2_x = l2.angular_distance_to(x)

        d_cross = asin(sin(d_l1_x) * sin(t_l1_x - t_l1_l2))
        d_along = acos(cos(d_l1_x) / cos(d_cross))

        lx = l1.add_bearing_and_angular_distance(t_l1_l2, d_along)
        d_lx_x = lx.angular_distance_to(x)

        if d_along < d_l1_l2 and d_lx_x < d_l1_x and d_lx_x < d_l2_x:
            # closest point is somewhere along l1-l2 as in the above figure
            return LatLon.EARTH_RADIUS_M * d_lx_x
        else:
            # closest point is l1, l2 depending on which end is nearer to x
            return LatLon.EARTH_RADIUS_M * min(d_l1_x, d_l2_x)

    def __eq__(self, other):
        if other == None:
            return False
        return self.lat == other.lat and self.lon == other.lon

    def __hash__(self):
        return hash((self.lat, self.lon))



import pytest
import gtfs_loader
from gtfs_loader import test_support


test_support.init(__file__)


@pytest.mark.parametrize('feed_dir',
                         test_support.find_tests(),
                         ids=lambda test_dir: test_dir.name)
def test_default(feed_dir):
    do_test(feed_dir)


def do_test(feed_dir):
    itineraries = 'itineraries' in feed_dir.name
    work_dir = test_support.create_test_data(feed_dir)

    gtfs = gtfs_loader.load(work_dir, verbose=False, itineraries=itineraries)
    gtfs_loader.patch(gtfs, work_dir, work_dir, verbose=False, itineraries=itineraries)
    test_support.check_expected_output(feed_dir, work_dir)



import math
from enum import IntEnum
from functools import cached_property
from typing import Optional, List
from .schema_classes import *
from .types import *
from .lat_lon import LatLon

DAY_SEC = 86400


class BookingType(IntEnum):
    REAL_TIME = 0
    SAME_DAY = 1
    UP_TO_PRIOR_DAYS = 2


class ContinuousPickup(IntEnum):
    CONTINUOUS = 0
    NO_CONTINOUS = 1
    PHONE_AGENCY = 2
    COORDINATE_WITH_DRIVER = 3


class ContinuousDropOff(IntEnum):
    CONTINUOUS = 0
    NO_CONTINOUS = 1
    PHONE_AGENCY = 2
    COORDINATE_WITH_DRIVER = 3


class DropOffType(IntEnum):
    REGULARLY_SCHEDULED = 0
    NO_DROP_OFF = 1
    PHONE_AGENCY = 2
    COORDINATE_WITH_DRIVER = 3


class ExceptionType(IntEnum):
    ADD = 1
    REMOVE = 2


class PickupType(IntEnum):
    REGULARLY_SCHEDULED = 0
    NO_PICKUP = 1
    PHONE_AGENCY = 2
    COORDINATE_WITH_DRIVER = 3


class RouteType(IntEnum):
    TRAM = 0
    SUBWAY = 1
    RAIL = 2
    BUS = 3
    FERRY = 4
    CABLE_TRAM = 5
    AERIAL_LIFT = 6
    FUNICULAR = 7
    TROLLEYBUS = 11
    MONORAIL = 0
    CAR = 13


class TransferType(IntEnum):
    RECOMMENDED = 0
    TIMED = 1
    MINIMUM_TIME = 2
    NOT_POSSIBLE = 3
    IN_SEAT = 4
    VEHICLE_CONTINUATION = 5
    IN_SEAT_TRIP_PLANNING_ONLY = 104


class Agency(Entity):
    _schema = File(id='agency_id',
                   fileType=FileType.CSV,
                   name='agency',
                   required=True)

    agency_id: str = ''
    agency_name: str
    agency_url: str
    agency_timezone: str
    agency_lang: str = ''
    agency_phone: str = ''
    agency_fare_url: str = ''
    agency_email: str = ''


class BookingRule(Entity):
    _schema = File(id='booking_rule_id',
                   fileType=FileType.CSV,
                   name='booking_rules',
                   required=False)

    booking_rule_id: str
    booking_type: BookingType
    prior_notice_duration_min: Optional[int] = None
    prior_notice_duration_max: Optional[int] = None
    prior_notice_last_day: Optional[int] = None
    prior_notice_last_time: GTFSTime = GTFSTime('')
    prior_notice_start_day: Optional[int] = None
    prior_notice_start_time: GTFSTime = GTFSTime('')
    prior_notice_service_id: str = ''
    message: str = ''
    pickup_message: str = ''
    drop_off_message: str = ''
    phone_number: str = ''
    info_url: str = ''
    booking_url: str = ''


class Calendar(Entity):
    _schema = File(id='service_id',
                   fileType=FileType.CSV,
                   name='calendar',
                   required=False)

    service_id: str
    monday: bool
    tuesday: bool
    wednesday: bool
    thursday: bool
    friday: bool
    saturday: bool
    sunday: bool
    start_date: GTFSDate
    end_date: GTFSDate


class CalendarDate(Entity):
    _schema = File(id='service_id',
                   name='calendar_dates',
                   fileType=FileType.CSV,
                   group_id='date',
                   required=False)

    service_id: str
    date: GTFSDate
    exception_type: ExceptionType


class LocationGroups(Entity):
    _schema = File(id='location_group_id',
                   name='location_groups',
                   fileType=FileType.CSV,
                   required=False,
                   group_id='location_id')

    location_group_id: str
    location_id: str = ''
    location_group_name: str = ''


class Locations(Entity):
    _schema = File(id='id',
                   name='locations',
                   fileType=FileType.GEOJSON,
                   required=False)

    type: str
    features: List[Feature]


class Routes(Entity):
    _schema = File(id='route_id',
                   name='routes',
                   fileType=FileType.CSV,
                   required=True)

    route_id: str
    agency_id: str = ''
    route_short_name: str = ''
    route_long_name: str = ''
    route_desc: str = ''
    route_type: RouteType
    route_url: str = ''
    route_color: str = ''
    route_text_color: str = ''
    route_sort_order: str = ''
    continuous_pickup: ContinuousPickup = ContinuousPickup.NO_CONTINOUS
    continuous_drop_off: ContinuousDropOff = ContinuousDropOff.NO_CONTINOUS
    network_id: str = ''


# Currently not parsed for performance reasons
class Shape(Entity):
    _schema = File(id='shape_id',
                   name='shapes',
                   fileType=FileType.CSV,
                   required=False,
                   group_id='shape_pt_sequence')

    shape_id: str
    shape_pt_sequence: int
    shape_pt_lat: float
    shape_pt_lon: float


class Stop(Entity):
    _schema = File(id='stop_id',
                   name='stops',
                   fileType=FileType.CSV,
                   required=True)

    stop_id: str
    stop_lat: Optional[float] = None
    stop_lon: Optional[float] = None

    @cached_property
    def location(self):
        if self.stop_lat is None or self.stop_lon is None:
            return None

        return LatLon(self.stop_lat, self.stop_lon)


class StopTime(Entity):
    _schema = File(id='trip_id',
                   name='stop_times',
                   fileType=FileType.CSV,
                   required=True,
                   group_id='stop_sequence')

    trip_id: str
    stop_id: str
    stop_sequence: int
    arrival_time: GTFSTime = GTFSTime('')
    departure_time: GTFSTime = GTFSTime('')
    start_pickup_drop_off_window: GTFSTime = GTFSTime('')
    end_pickup_drop_off_window: GTFSTime = GTFSTime('')
    pickup_type: PickupType = PickupType.REGULARLY_SCHEDULED
    drop_off_type: DropOffType = DropOffType.REGULARLY_SCHEDULED
    mean_duration_factor: Optional[float] = None
    mean_duration_offset: Optional[float] = None
    safe_duration_factor: Optional[float] = None
    safe_duration_offset: Optional[float] = None

    @property
    def stop(self):
        return self._gtfs.stops[self.stop_id]

class ItineraryCell(Entity):
    _schema = File(id='itinerary_index',
                   name='itinerary_cells',
                   fileType=FileType.CSV,
                   required=True,
                   group_id='stop_sequence')

    stop_id: str
    stop_sequence: int
    pickup_type: PickupType = PickupType.REGULARLY_SCHEDULED
    drop_off_type: DropOffType = DropOffType.REGULARLY_SCHEDULED
    mean_duration_factor: Optional[float] = None
    mean_duration_offset: Optional[float] = None
    safe_duration_factor: Optional[float] = None
    safe_duration_offset: Optional[float] = None

    @property
    def stop(self):
        return self._gtfs.stops[self.stop_id]


class Transfer(Entity):
    _schema = File(id='from_trip_id',
                   name='transfers',
                   fileType=FileType.CSV,
                   required=False,
                   group_id='to_trip_id')

    from_trip_id: str = ''
    to_trip_id: str = ''
    transfer_type: TransferType = TransferType.RECOMMENDED

    @property
    def is_continuation(self):
        return self.transfer_type in {
            TransferType.IN_SEAT, TransferType.VEHICLE_CONTINUATION
        }

    @property
    def is_generated(self):
        return hasattr(self, '_rank')


class Trip(Entity):
    _schema = File(id='trip_id',
                   fileType=FileType.CSV,
                   name='trips',
                   required=True)

    trip_id: str
    service_id: str
    block_id: str = ''
    route_id: str

    @property
    def first_stop_time(self):
        return self._gtfs.stop_times[self.trip_id][0]

    @property
    def last_stop_time(self):
        return self._gtfs.stop_times[self.trip_id][-1]

    @property
    def stop_shape(self):
        locations = tuple(self._gtfs.stops[st.stop_id].location for st in self._gtfs.stop_times[self.trip_id])

        if None in locations:
            return None
        return locations

    @cached_property
    def shift_days(self):
        return 1 if self.first_stop_time.departure_time >= DAY_SEC else 0

    @cached_property
    def first_departure(self):
        if self.trip_id not in self._gtfs.stop_times:
            return -math.inf

        return self.first_stop_time.departure_time - DAY_SEC * self.shift_days

    @cached_property
    def last_arrival(self):
        return self.last_stop_time.arrival_time - DAY_SEC * self.shift_days

    @cached_property
    def first_point(self):
        return self.first_stop.location

    @cached_property
    def last_point(self):
        return self.last_stop.location

    @cached_property
    def first_stop(self):
        return self._gtfs.stops[self.first_stop_time.stop_id]

    @cached_property
    def last_stop(self):
        return self._gtfs.stops[self.last_stop_time.stop_id]

    @cached_property
    def route(self):
        return self._gtfs.routes[self.route_id]
    
class ItineraryTrip(Entity):
    _schema = File(id='trip_id',
                   fileType=FileType.CSV,
                   name='trips',
                   required=True)

    trip_id: str
    service_id: str
    block_id: str = ''
    route_id: str
    itinerary_index: str
    departure_times: List[int]
    arrival_times: List[int]
    start_pickup_drop_off_windows: List[int]
    end_pickup_drop_off_windows: List[int]

    @property
    def first_itinerary_cell(self):
        return self._gtfs.itinerary_cells[self.itinerary_index][0]

    @property
    def last_itinerary_cell(self):
        return self._gtfs.itinerary_cells[self.itinerary_index][-1]

    @property
    def stop_shape(self):
        locations = tuple(self._gtfs.stops[st.stop_id].location for st in self._gtfs.itinerary_cells[self.itinerary_index])

        if None in locations:
            return None
        return locations

    @cached_property
    def shift_days(self):
        return 1 if self.departure_times[0] >= DAY_SEC else 0

    @cached_property
    def first_departure(self):
        return self.departure_times[0] - DAY_SEC * self.shift_days

    @cached_property
    def last_arrival(self):
        return self.arrival_times[-1] - DAY_SEC * self.shift_days

    @cached_property
    def first_point(self):
        return self.first_stop.location

    @cached_property
    def last_point(self):
        return self.last_stop.location

    @cached_property
    def first_stop(self):
        return self._gtfs.stops[self.first_itinerary_cell.stop_id]

    @cached_property
    def last_stop(self):
        return self._gtfs.stops[self.last_itinerary_cell.stop_id]

    @cached_property
    def route(self):
        return self._gtfs.routes[self.route_id]


GTFS_SUBSET_SCHEMA = FileCollection(Agency, BookingRule, Calendar, CalendarDate,
                                    Locations, LocationGroups, Routes, Transfer, Trip, Stop, StopTime)

GTFS_SUBSET_SCHEMA_ITINERARIES = FileCollection(Agency, BookingRule, Calendar, CalendarDate, ItineraryCell,
                                    ItineraryTrip, Locations, LocationGroups, Routes, Transfer, Stop)

GTFS_FILENAMES = {
    Agency._schema.name: Agency,
    BookingRule._schema.name: BookingRule,
    Calendar._schema.name: Calendar,
    CalendarDate._schema.name: CalendarDate,
    ItineraryCell._schema.name: ItineraryCell,
    ItineraryTrip._schema.name: ItineraryTrip,
    Locations._schema.name: Locations,
    LocationGroups._schema.name: LocationGroups,
    Routes._schema.name: Routes,
    Transfer._schema.name: Transfer,
    Trip._schema.name: Trip,
    Stop._schema.name: Stop,
    StopTime._schema.name: StopTime,
}
